
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Paper: H. Jia and Y.M. Cheung,Subspace Clustering of Categorical and Numerical
%%%% Data With an Unknown Number of Clusters, IEEE Transactions on Neural Networks 
%%%% and Learning Systems, vol. 29, no. 8, pp. 3308-3325, 2018.  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% Code of Attribute-weighted OCIL algorithm (WOCIL) with Oriented initialization 
%%%%%%  for categorical data clustering %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

close all;
clear all;

%
xx = textread('./new_soybean_small.txt');  %%% input data sets
x = xx(:,1:35);
true_class = xx(:,36);
true_class = true_class';
x = x + 1;
[n, d] = size(x);  %%% n: the number of samples;   d: the number of attributes(dimensionality) 

k = 4;  %%% number of clusters
m = 10; %%% the maximum number of different values can be chosen by each attribute  
%}


T_total = 50; %%%% Repeat 50 times
ACC_record = zeros(1, T_total);
RI_record = zeros(1, T_total);
NMI_record = zeros(1, T_total);

for T_times = 1:T_total
    
%     
tic

sample_frequency = zeros(d, m);
for i = 1:n
    for j = 1:d
        sample_frequency(j, x(i, j)) = sample_frequency(j, x(i, j)) + 1;
    end
end

feature_weight = ones(d, k)./d;
miu = zeros(k, d);  %%% k modes

    %%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%% Initialization method 2: Oriented initialization method %%%%%%
    %
    value_frequency = sample_frequency';
    first_select_proba = zeros(n, d);
    for i = 1:n
        for j = 1:d
            first_select_proba(i, j) = value_frequency(x(i,j),j)./n;
        end
    end
    first_select_proba = first_select_proba';
    data_set_similar = sum(first_select_proba)./d;

    [~, first_select_index] = max(data_set_similar);
    miu(1, :) = x(first_select_index, :);

    frequency_in_miu = zeros(m, d);
    proba_select_value = zeros(1, d);
    similary_miu = zeros(1, n);
    
    for j = 1:d
        frequency_in_miu(miu(1, j),j) = frequency_in_miu(miu(1, j),j) + 1;
    end
    
    miu_num = 1;
    while (miu_num < k)
        for i = 1:n
            for j = 1:d
                proba_select_value(j) = frequency_in_miu(x(i, j),j)/miu_num;
            end
            similary_miu(i) = sum(proba_select_value)/d; 
        end
    
        select_criterion = data_set_similar + (1 - similary_miu); 
        [~, max_index] = max(select_criterion);
        miu_num = miu_num + 1;
        miu(miu_num, :) = x(max_index, :);
    
        for j = 1:d
            frequency_in_miu(miu(miu_num, j),j) = frequency_in_miu(miu(miu_num, j),j) + 1;
        end
    end
    %}
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
%%
frequency_in_cluster = zeros(d, m, k);  %%%%count the frequency of each attribute value amoug the initial classes
for c = 1:k
    for j = 1:d
        frequency_in_cluster(j,miu(c,j),c) = frequency_in_cluster(j,miu(c,j),c)+1;
    end
end

%%%%% Assign the data into the initial k clusters %%%%%%%%%%
cluster_label = zeros(1, n);
cluster_count = ones(1, k);
for i = 1:n
    class_label = Weighted_Class_assign(x(i,:), frequency_in_cluster, cluster_count, feature_weight);
    cluster_label(i) = class_label;
    cluster_count(class_label) = cluster_count(class_label)+1;
    for j = 1:d
        frequency_in_cluster(j, x(i,j), class_label) = frequency_in_cluster(j, x(i,j), class_label)+1;
    end
end

frequency_out_cluster = zeros(d, m, k);
feature_weight_value = zeros(d, k);

%%%%%%%% Iterative steps: adjust the cluster label of each data untill no data need to change class. %%%%%%%%%%%%%%%%%%%%%%%%%%
moves = 1;
epoch = 0;
while (moves ~= 0) %%% some class label has been changed in the last step%%%
    moves = 0;
    
    %%%%%%%%%%%%%%%% update feature weights %%%%%%%%%%%%%%%%%%%
    for c = 1:k
        frequency_out_cluster(:, :, c) = sample_frequency - frequency_in_cluster(:, :, c);
    end
    for j = 1:d
        for c = 1:k
            feature_weight_Fval = 0;
            feature_weight_Mval = 0;
            for t = 1:m
                feature_weight_Fval = feature_weight_Fval + (frequency_in_cluster(j, t, c)/cluster_count(c)-frequency_out_cluster(j, t, c)/(n-cluster_count(c)))^2;
                feature_weight_Mval = feature_weight_Mval + frequency_in_cluster(j, t, c)*(frequency_in_cluster(j, t, c)/cluster_count(c));
            end
            feature_weight_F = sqrt(feature_weight_Fval)/(sqrt(2));
            feature_weight_M = feature_weight_Mval/cluster_count(c);
            feature_weight_value(j, c) = feature_weight_F * feature_weight_M;
        end
    end
    weight_regulize = sum(feature_weight_value);
    for c = 1:k
        feature_weight(:, c) = feature_weight_value(:, c)./weight_regulize(c);
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%% update cluster membership %%%%%%%%%%%%%%%%%%%
    for i = 1:n        
        new_label = Weighted_Class_assign(x(i,:), frequency_in_cluster, cluster_count, feature_weight);

        if cluster_label(i) ~= new_label  %%% need adjust
            moves = moves + 1;
            old_cluster = cluster_label(i);
            cluster_label(i) = new_label;     %%%%updata cluster label
            cluster_count(new_label) = cluster_count(new_label) + 1;
            cluster_count(old_cluster) = cluster_count(old_cluster) - 1;   %%%%update cluster counter
            for j = 1:d        %%%%update the modes of the two clusters
                frequency_in_cluster(j, x(i,j), new_label) = frequency_in_cluster(j, x(i,j), new_label) + 1;       
                frequency_in_cluster(j, x(i,j), old_cluster) = frequency_in_cluster(j, x(i,j), old_cluster) - 1; 
            end
        end
    end
    epoch = epoch + 1;
end

exetime = toc;

%%
%%%% Clustering error
[accuracy,Confu_Matrix,~,~] = calAccuracy(cluster_label',true_class');

%%
%%% RI index
RI_value = calRI(cluster_label,true_class);

%%
%%% NMI index
[NMI_value,Enlarge_Matrix, I_index,H_obtain,H_true] = calNMI(Confu_Matrix, n);

%%%%%%%%%
ACC_record(T_times) = accuracy;
RI_record(T_times) = RI_value;
NMI_record(T_times) = NMI_value;
end

ACC_mean = mean(ACC_record)
ACC_std = std(ACC_record)

RI_mean = mean(RI_record)
RI_std = std(RI_record)

NMI_mean = mean(NMI_record)
NMI_std = std(NMI_record)
